#include	"defs.h"
#include	"input.h"
#include	"print.h"
#include	"string.h"
#include	"name.h"
#include	"alloc.h"

#define	SELF	(char *)0

typedef	unsigned char	pschar;

typedef struct {
	char	*name;
	pschar	code;
	char	*font;
} nchar;

typedef struct C {
	nchar	uch;
	struct C *next;
} uchar;

static	uchar	*uchars = (uchar *)0;

static	nchar	namedchars[] = {
	"bullet",	(pschar)'\267',		SELF,	/* bullet */
	"emdash",	(pschar)'\320',		SELF,	/* em dash */
	"oe",		(pschar)'\372',		SELF,	/* oe diphthong */
	"dagger",	(pschar)'\262',		SELF,	/* dagger */
	"ddagger",	(pschar)'\263',		SELF,	/* double dagger */
	"minus",	(pschar)'\261',		SELF,	/* minus sign */
	"section",	(pschar)'\247',		SELF,	/* section symbol */
	"paragraph",	(pschar)'\266',		SELF,	/* paragraph symbol */
	"pound",	(pschar)'\243',		SELF,	/* pound sign */
	"dquote",	(pschar)'\042',		SELF,	/* double quote */
	"oquote",	(pschar)'\252',		SELF,	/* open quote */
	"cquote",	(pschar)'\272',		SELF,	/* close quote */
	"co",		(pschar)'\323',		"S",	/* copyright */
	"rg",		(pschar)'\322',		"S",	/* registered */
	"left",		(pschar)'\254',		"S",	/* left arrow */
	"up",		(pschar)'\255',		"S",	/* up arrow */
	"right",	(pschar)'\256',		"S",	/* right arrow */
	"down",		(pschar)'\257',		"S",	/* down arrow */
	"plusmin",	(pschar)'\261',		"S",	/* plus or minus */
	"return",	(pschar)'\277',		"S",	/* carriage return */
	"square",	(pschar)'\156',		"ZD",	/* black square */
	"rhand",	(pschar)'\053',		"ZD",	/* right hand */
	"phone",	(pschar)'\045',		"ZD",	/* telephone */
	"scissor",	(pschar)'\042',		"ZD",	/* scissors */
	(char *)0
};

static	nchar	*lookup_userchar();


void set_char()			/* .ch name val font */
{
	register uchar *up;
	nchar *np;
	char *name, *ft, c;
	int val = 0;
	extern char *lookup_font();

	if (peeknb() == EOS)
		RET_WARN("no arguments given to '.ch'");
	name = mk_string(trim(getwd()));
	if (peeknb() == EOS)
		RET_WARN("no character value given with '.ch'");
	c = getnb();
	while (c >= '0' && c <= '7') {
		val = (val << 3) | (c - '0');
		c = getch();
	}
	if (val < 0 || val > 255)
		RET_WARN("bad character value given to '.ch'");
	if (peeknb() != EOS) {
		ft = lookup_font(trim(getwd()));
		if (ft == (char *)0)
			warn("bad font name given to '.ch'");
	} else
		ft = SELF;
	if ((np = lookup_userchar(name)) == (nchar *)0) {
		up = (uchar *)alloc(sizeof(*up), "user character");
		up->uch.name = name;
		up->uch.code = (unsigned char)val;
		up->uch.font = ft;
		up->next = uchars;
		uchars = up;
	} else {
		np->code = (unsigned char)val;
		np->font = ft;
	}
}

void rm_char()			/* .rc name ... */
{
	register uchar *up, *uq;
	char *name;

	if (peeknb() == EOS)
		RET_WARN("no argument given to '.rc'");
	do {
		name = trim(getwd());
		for (up = uchars, uq = (uchar *)0; up; uq = up, up = up->next)
			if (streq(up->uch.name, name))
				break;
		if (up == (uchar *)0) {
			warn("non-existent character name given to '.rc'");
			continue;
		}
		if (uq == (uchar *)0)
			uchars = up->next;	/* first on list */
		else
			uq->next = up->next;
		free(up->uch.name);
		free((char *)up);
	} while (peeknb() != EOS);
}

static nchar *lookup_userchar(name)
char *name;
{
	register uchar *up;

	for (up = uchars; up; up = up->next)
		if (streq(up->uch.name, name))
			return &up->uch;
	return (nchar *)0;
}

char *lookup_char(name)
register char *name;
{
	register nchar *cp;
	static char ch[8], fch[1024];
	extern int current_ps();
	extern char *current_font();

	cp = lookup_userchar(name);
	if (cp == (nchar *)0) {
		register bool abb;
		
		abb = strlen(name) == 2 ? T : F;
		for (cp = namedchars; cp->name; cp++)
			if ((abb == T && strneq(cp->name, name, 2)) ||
			    (abb == F && streq(cp->name, name)))
				break;
	}
	if (cp->name == (char *)0) {
		warn("interpolation of undefined character");
		return (char *)0;
	}
	if (cp->code == '\0')
		return (char *)0;
	(void) sprintf(ch, "(\\%03o)", cp->code);
	if (cp->font == SELF)
		return ch;
	(void) sprintf(fch, "[%d %s]%s%s",
		current_ps(), cp->font, ch, current_font());
	return fch;
}
