#include	"defs.h"
#include	"input.h"
#include	"line.h"
#include	"scalenum.h"
#include	"print.h"
#include	"string.h"
#include	"alloc.h"

#define	STRLEN	BUFSIZE

typedef struct I {
	line	*lines;		/* the text */
	struct I *prev;
} ifobj;

static	ifobj	*ifs = (ifobj *)0;	/* stack index */
extern	bool	expand;

static	void	do_conditional();
static	void	do_statement();
static	int	do_string_cmp();
static	int	do_num_expr();
static	bool	read_string();
static	bool	advance();
static	void	mk_line();


void do_if()			/* .if <cond> <state> */
{
	if (peeknb() == EOS)
		RET_WARN("no condition given to '.if'");
	do_conditional(F);
}

void do_ifelse()		/* .ifel <cond> <state>\n.else <state> */
{
	if (peeknb() == EOS)
		RET_WARN("no condition given to '.ifel'");
	do_conditional(T);
}

static void do_conditional(ifelse)
bool ifelse;
{
	char c;
	int expr_true;
	bool cond, not = F;
	itext i;

	/* read and evaluate the test expression */
	if ((c = getnb()) == '!') {
		not = T;
		c = getch();
	}
	ungetch(c);
	if (strchr("+-0123456789.(", c))
		expr_true = do_num_expr();
	else
		expr_true = do_string_cmp();
	if (expr_true < 0)	/* error in test expression */
		return;
	cond = expr_true ? T : F;
	if (not == T)
		cond = NOT(cond);

	/* handle the 'if' part */
	do_statement(cond, advance());
	if (ifelse == T) {
		/* handle the 'else' part */
		if (readline("-> ") == F ||
		    ((c = getch()) != '.' && c != '\'') ||
		    !(strneq(getwd(), "else", 4)))
			RET_WARN("missing .else with .ifel");
		do_statement(NOT(cond), advance());
	}
	if (ifelse == T || cond == T) {
		/* push back the if or else text to be executed */
		i.lines = ifs->lines;
		pushinline(i, "if", 0);
	}
}


static int do_num_expr()
{
	int res = 0;

	scale_num(1, T);
	if (num.err == T)
		return -1;
	set_num(&res, ANY);
	return res > 0;
}

static int do_string_cmp()
{
	char delim, str1[STRLEN], str2[STRLEN];

	delim = getnb();
	if (read_string(str1, delim) == F || read_string(str2, delim) == F)
		return -1;
	return streq(str1, str2);
}

static bool read_string(buf, delim)
char buf[], delim;
{
	register char *sp, *ep;

	for (sp = buf, ep = &buf[STRLEN]; (*sp = getch()) != delim; sp++)
		if (*sp == EOS || sp >= ep) {
			warn("bad string comparison in '.if' condition");
			return F;
		}
	*sp = EOS;
	return T;
}

static bool advance()
{
	char c;
	bool state = T;

	expand = F;
	if ((c = getnb()) == '\\' && peekch() == '{') {
		(void) getch();
		state = F;
	} else
		ungetch(c);
	expand = T;
	return state;
}

static void do_statement(collect, state)
bool collect, state;
{
	expand = F;
	if (collect == T) {
		/* save the statement/block for execution */
		ifobj *ip;

		ip = (ifobj *)alloc(sizeof(*ip), "'if' statement");
		ip->lines = (line *)0;
		ip->prev = ifs;
		ifs = ip;
	}
	if (state == T) {
		if (collect == T) {
			ungetch(getnb());
			mk_line(getln(REST));
		}
	} else {
		int nest = 1;
		char linebuf[4096], c;
		register char *lbp = linebuf;

		while (nest) {
			switch (*lbp = getch()) {
			case EOS:
				if (collect == T)
					mk_line(linebuf);
				lbp = linebuf;
				if (readline("-> ") == F)
					nest = 0;
				continue;
			case '\\':
				c = peekch();
				if (c == '{')
					nest++;
				else if (c == '}' && --nest == 0) {
					*lbp = EOS;
					if (collect == T)
						mk_line(linebuf);
					continue;
				}
				break;
			}
			lbp++;
		}
		(void) getch();
	}
	if (collect == T)
		mk_line(".PIf");
	expand = T;
}

static void mk_line(text)
char *text;
{
	line *lp;
	static line *lq;

	lp = (line *)alloc(sizeof(*lp), "'if' text");
	lp->text = mk_string(text);
	lp->next = (line *)0;
	if (ifs->lines == (line *)0)
		lq = ifs->lines = lp;
	else
		lq = lq->next = lp;
}

void pop_if()
{
	ifobj *ip = ifs->prev;
	line *lp, *lq;

	for (lp = ifs->lines; lp; lp = lq) {
		lq = lp->next;
		free(lp->text);
		free((char *)lp);
	}
	free((char *)ifs);
	ifs = ip;
}

void bad_else()
{
	warn("unexpected '.else' request");
}
