#include	<stdio.h>
#include	"defs.h"
#include	"input.h"
#include	"line.h"
#include	"print.h"
#include	"string.h"
#include	"alloc.h"

#define	LINESIZE	4096
#define	WORDSIZE	128

#define	WHITE(c)	((c) == ' ' || (c) == '\t')


static	char	textline[LINESIZE];	/* input line */
static	char	*p;			/* line ptr */
static	char	*linep = textline;	/* ptr to start of input line */
static	char	*savep;			/* save ptr for temp strings */
bool	expand = T;			/* false for no \ expansion */

static	int	linecnt;		/* line cnt in current stream */

typedef struct In {
	bool	file;			/* false if input from macro */
	bool	terminal;		/* true if input from terminal */
	union {
		FILE	*stream;
		line	*memtext;
	} s_m;
	char	*name;			/* file or macro name */
	int	linecount;		/* input line (when stacked) */
	struct In *last;
} input;				/* input stack element */
static	input	*in = (input *)0;	/* stack pointer */

static	void	stack_input();
static	bool	popinfile();
static	void	interp_ch();


bool readline(prompt)
char	*prompt;
{
	bool gotline = F;
	
	if (in == (input *)0)
		return F;
	if (in->file == T) {
		if (in->terminal == T)
			(void) fprintf(stderr, prompt);
		if ((p = fgets(textline, LINESIZE-1, in->s_m.stream)) != NULL) {
			int lastch = strlen(textline) - 1;
	
			if (textline[lastch] == '\n')
				textline[lastch] = EOS;	/* zap the NL */
			gotline = T;
		}
	} else {
		if (in->s_m.memtext) {
			register line *lp = in->s_m.memtext;
	
			in->s_m.memtext = lp->next;
			p = strcpy(textline, lp->text);
			free((char *)lp);
			gotline = T;
		}
	}
	if (gotline == T) {
		linecnt++;
		return T;
	}
	if (popinfile() == F)
		return F;
	return readline(prompt);
}


static void interp_ch()
{
	extern bool copymode, do_escape_sequ;
	extern void interp_reg(), interp_str(), interp_arg();

	if (*p == '\\' && do_escape_sequ == T && expand == T) {
		switch (p[1]) {
		case 'n':
			p += 2;
			interp_reg();
			break;
		case '*':
			p += 2;
			interp_str();
			break;
		case '$':
			p += 2;
			interp_arg();
			break;
		case '\\':
			if (copymode == T)
				p++;
			else
				p[1] = 'e';
			break;
		case '"':
			*p = EOS;
			break;
		case EOS:
			if (copymode == F) {
				if (readline("-> ") == F)
					*p = EOS;
				else
					interp_ch();
			}
			break;
		}
	}
}

char getch()
{
	if (*p == EOS)
		return EOS;
	interp_ch();
	return *p++;
}

char getnb()
{
	register char c;
	
	while ((c = getch()), WHITE(c));
	return c;
}

char peekch()
{
	interp_ch();
	return *p;
}

char peeknb()
{
	register char c;
	
	loop {
		c = peekch();
		if (!WHITE(c))
			break;
		(void) getch();
	}
	return c;
}

void ungetch(c)
char	c;
{
	if (p > linep)
		*--p = c;
}

char *getwd()
{
	static char wordbuf[WORDSIZE];
	register char *wp = wordbuf;
	
	*wp = getnb();
	do {
		*++wp = getch();
	} while (*wp && !WHITE(*wp) && wp < &wordbuf[WORDSIZE-1]);
	if (WHITE(*wp))
		ungetch(*wp);
	*wp = EOS;
	return wordbuf;
}

char *getln(howmuch)
{
	return howmuch == REST ? p : textline;
}


void pushtempstr(s)
char *s;
{
	savep = p;
	p = linep = s;
	expand = F;
}

void poptempstr()
{
	p = savep;
	linep = textline;
	expand = T;
}

void pushinstr(str)
char *str;
{
	register strlent llen, slen;
	register char *q;

	if ((llen = strlen(textline)) + (slen = strlen(str)) >= LINESIZE)
		ferr("input overflow after reg/string/macro expansion\n", "");
	q = &textline[llen + slen];
	llen = strlen(p) + 1; 
	for ( ; llen--; q--)
		*q = q[-slen];
	(void) strncpy(p, str, slen);
}

void pushinline(l, name, str)
itext l;
char *name;
{
	register line *lp, *lq, *clp;
	line *lfirst = (line *)0;

	if (!str)
		clp = l.lines;
	do {
		lp = (line *)alloc(sizeof(*lp), "text");
		if (str) {
			(lfirst = lp)->text = l.string;
			break;
		}
		lp->text = clp->text;
		if (lfirst == (line *)0)
			lq = lfirst = lp;
		else
			lq = lq->next = lp;
		clp = clp->next;
	} while (clp);
	lp->next = (line *)0;
	stack_input(name);
	in->file = in->terminal = F;
	in->s_m.memtext = lfirst;
}

void pushinfile(stream, name)
FILE	*stream;
char	*name;
{
	extern int isatty();

	stack_input(name);
	in->file = T;
	in->terminal = isatty(fileno(stream)) ? T : F;
	in->s_m.stream = stream;
}

static void stack_input(name)
char *name;
{
	input *ip;

	ip = (input *)alloc(sizeof(*ip), "input structure");
	if (in)
		in->linecount = linecnt;
	ip->last = in;
	in = ip;
	in->name = mk_string(name);
	linecnt = 0;
}

static bool popinfile()
{
	input *ip;

	if (in == (input *)0)
		return F;
	free(in->name);
	if (in->file == T && in->terminal == F && fclose(in->s_m.stream) == EOF)
		return F;
	ip = in->last;
	free((char *)in);
	if ((in = ip) == (input *)0)
		return F;
	linecnt = in->linecount;
	return T;
}


void errposn(mess)
char	*mess;
{
	input *ip;
	int i = 1, n;

	if (in->terminal == T)
		(void) fputs("--- ", stderr);
	else
		(void) fprintf(stderr, "%s [%d] ", in->name, linecnt);
	(void) fprintf(stderr, "%s\n", mess);
	for (ip = in->last; ip && ip->terminal == F; ip = ip->last) {
		for (n = i++; n--; )
			(void) fputs("  ", stderr);
		(void) fprintf(stderr, "<- %s [%d]\n", ip->name, ip->linecount);
	}
}
