#include	"defs.h"
#include	"input.h"
#include	"regfmts.h"
#include	"interpol.h"
#include	"scalenum.h"
#include	"string.h"
#include	"output.h"
#include	"print.h"


/* LIST ENVIRONMENTS (NESTED) */
#define	MAXNEST	4

#define	LIST(x)	list[llev].x

static struct {
	int	type;		/* list type (see defines below) */
	char	*mark;		/* item mark string */
	int	mval;		/* current mark value for ALPNUM lists */
	int	p_in;		/* pre-list indent (restored on .le) */
	int	t_in;		/* text indent */
	int	m_in;		/* mark indent */
	int	sp;		/* list spacing (0, 1 or 2) */
} list[MAXNEST];

static	int	llev = -1;	/* current list level */

/* DEFAULT LIST VALUES */
/* types 1 to 13 are in regfmts.h, i.e. register formats */
#define	BULLET	14
#define	DASH	15
#define	LIMARK	16

#define	DFLTVAL(x)	dflt_list[LIST(type)-1].x

static struct {
	int	t_in;		/* text indent */
	int	m_in;		/* mark indent */
} dflt_list[] = {
	{ 36, 14 }, { 36,  7 }, { 36,  0 }, {  0,  0 },
	{  0,  0 }, {  0,  0 }, {  0,  0 }, {  0,  0 },
	{  0,  0 }, { 36,  0 }, { 36,  0 }, { 36, 14 },
	{ 36, 14 }, { 24,  7 }, { 24,  0 }, { 36,  0 }
};

/* FUNCTION DECLARATIONS */
static	void	mk_mark();	/* prepare and set mark and type */
extern	void	do_changes();	/* do any changes asked for */
extern	void	set_list_indent(); /* set indent for list */
extern	int	current_ps();	/* get current point size */
extern	int	current_vs();	/* get current vert spacing */
extern	int	current_in();	/* get current indentation */


void set_list()			/* .ls [mk] [t_in] [m_in] [sp] */
{
	if (++llev == MAXNEST)
		RET_WARN("too many nested lists");
	LIST(mval) = 1;
	LIST(sp) = 1;
	if (peeknb() == EOS)
		mk_mark("a)");
	else {
		get_string();
		mk_mark(string);
	}
	LIST(p_in) = current_in();
	if (LIST(type) >= 4 && LIST(type) <= 9) /* arabic 4 to 9 digits */
		LIST(t_in) = LIST(type) * (current_ps() / 2) + 14;
	else
		LIST(t_in) = DFLTVAL(t_in);
	LIST(m_in) = DFLTVAL(m_in);
	scale_num(72, F);
	if (num.err == F)
		set_num(&LIST(t_in), 5);
	scale_num(72, F);
	if (num.err == F)
		set_num(&LIST(m_in), 0);
	scale_num(1, F);
	if (num.err == F)
		set_num(&LIST(sp), 0);
	do_changes();
	out_br();
	if (LIST(sp) > 1)
		out_sp(current_vs() / 2);
	set_list_indent(LIST(p_in) + LIST(t_in));
}

void end_list()			/* .le [sp] */
{
	bool arg = F;

	if (llev < 0)
		RET_WARN("no current list to end");
	if (getch() != EOS)
		arg = T;
	do_changes();
	out_br();
	if (arg == T || LIST(sp) > 1)
		out_sp(current_vs() / 2);
	set_list_indent(LIST(p_in));
	llev--;
}

void do_list_item()		/* .li [mark] */
{
	char mbuf[1024], *mp = mbuf;
	bool nl = F;
	extern char *format_reg();
	extern int abs_x();

	if (llev < 0)
		RET_WARN("not in a list");
	do_changes();
	out_br();
	if (LIST(sp) == 1)
		out_sp(current_vs() / 2);
	out_mv(abs_x(LIST(p_in) + LIST(m_in)));
	if (peeknb() == EOS) {
		if (LIST(type) == LIMARK)
			return;		/* hanging indent */
		if (LIST(type) == BULLET || LIST(type) == DASH)
			(void) sprintf(mbuf, LIST(mark));
		else
			(void) sprintf(mbuf, LIST(mark),
				format_reg(LIST(mval)++, LIST(type)));
	} else {
		get_string();
		if (LIST(type) == LIMARK)
			(void) sprintf(mbuf, LIST(mark), string);
		else
			mp = string;
		mp = interpolate(mp, NULLSTR, NULLSTR, 1);
	}
	out_line(mp, nl);
	out_imv();
}

static void mk_mark(s)
char *s;
{
	char buf[256], *bp = buf;
	bool found = F;
#define	addstr(s)	{ *bp = EOS; bp = buf + strlen(strcat(buf, s)); }

	for ( ; *s; s++) {
		if (found == F) {
			switch (*s) {
			case 'a':
				LIST(type) = ALPHA; break;
			case 'A':
				LIST(type) = ALPHACAPS; break;
			case '1':
				LIST(type) = ARABIC; break;
			case '2': case '3': case '4': case '5':
			case '6': case '7': case '8': case '9':
				LIST(type) = *s - '0'; break;
			case 'i':
				LIST(type) = ROMAN; break;
			case 'I':
				LIST(type) = ROMANCAPS; break;
			case 'b':
				LIST(type) = BULLET;
				addstr("\\267");
				break;
			case 'd':
				LIST(type) = DASH;
				addstr("\\320");
				break;
			case 'v':
				LIST(type) = LIMARK;
				break;
			default:
				*bp++ = *s;
				continue;
			}
			found = T;
			if (LIST(type) != BULLET && LIST(type) != DASH)
				addstr("%s");
		} else
			*bp++ = *s;
	}
	*bp = EOS;
	if (LIST(type) == LIMARK)
		bp = buf;
	else
		bp = interpolate(buf, NULLSTR, NULLSTR, 1);
	LIST(mark) = mk_string(bp);
}
