#include	"defs.h"
#include	"input.h"
#include	"scalenum.h"
#include	"regfmts.h"
#include	"fmtregs.h"
#include	"name.h"
#include	"print.h"
#include	"string.h"
#include	"alloc.h"

typedef struct N {
	char	*name;
	int	value;
	int	incr;
	int	fmt;	/* see above defines; 2<=fmt<=9 implies %0<fmt>d */
	struct N *next;
} reg;

static	reg	*regs = (reg *)0;

char	*format_reg();
static	reg	*lookup();
static	reg	*getreg();
static	reg	*addreg();


void init_register(s)		/* initialise register from cmdline arg. */
register char *s;
{
	register char *vp;
	register reg *rp;
	extern int atoi();

	vp = strrchr(s, '=');
	if (vp && vp > s)
		if (vp[-1] == '\\')
			vp = NULLSTR;	/* no assignment value */
		else *vp++ = EOS;
	rp = getreg(trim(s));
	if (vp)
		rp->value = atoi(vp);
}

void set_ireg(name, val)	/* set internal register */
char *name;
{
	getreg(name)->value = val;
}

void set_register()		/* .nr reg [val] [incr] */
{
	register reg *rp;

	if (peeknb() == EOS)
		RET_WARN("no arguments given to '.nr'");
	rp = getreg(trim(getwd()));
	if (peeknb() == EOS)
		return;
	scale_num(1, F);
	if (num.err == F)
		set_num(&rp->value, ANY);
	if (peeknb() == EOS)
		return;
	scale_num(1, T);
	if (num.err == T)
		RET_WARN("non-numeric increment value given to '.nr'");
	set_num(&rp->incr, ANY);
}

void set_reg_format()		/* .af reg fmt */
{
	register reg *rp;
	char c;

	if (peeknb() == EOS)
		RET_WARN("no arguments given to '.af'");
	rp = getreg(trim(getwd()));
	if (peeknb() == EOS)
		RET_WARN("no format given to '.af'");
	switch (c = getnb()) {
	case '1':	rp->fmt = ARABIC; break;
	case 'i':	rp->fmt = ROMAN; break;
	case 'I':	rp->fmt = ROMANCAPS; break;
	case 'a':	rp->fmt = ALPHA; break;
	case 'A':	rp->fmt = ALPHACAPS; break;
	default:	if (c >= '2' && c <= '9')
				rp->fmt = c - '0';
			else
				RET_WARN("unknown format given to '.af'");
	}
}

void rm_register()		/* .rr reg ... */
{
	register reg *rp, *rq;

	if (peeknb() == EOS)
		RET_WARN("no argument given to '.rr'");
	do {
		if ((rp = lookup(trim(getwd()))) == (reg *)0) {
			warn("non-existent register name given to '.rr'");
			continue;
		}
		for (rq = regs; rq; rq = rq->next)
			if (rq->next == rp)
				break;
		if (rq == (reg *)0)	/* reg is first on list */
			regs = rp->next;
		else
			rq->next = rp->next;
		free(rp->name);
		free((char *)rp);
	} while (peeknb() != EOS);
}

void interp_reg()		/* \n[+-]r or \n[+-](rr or \n[+-]{reg} */
{
	char c;
	int incr = 0, n = 1;
	reg *rp;

	c = getch();
	if (c == '+' || c == '-') {
		incr = c == '+' ? 1 : -1;
		c = getch();
	}
	if (c == '(' || c == '{')
		n = c == '(' ? 2 : 0;
	else
		ungetch(c);
	rp = lookup(trim(getname(n)));
	if (rp == (reg *)0) {
		pushinstr("0");
		RET_WARN("interpolation of undefined number register")
	}
	if (incr)
		rp->value += incr * rp->incr;
	pushinstr(format_reg(rp->value, rp->fmt));
}

char *format_reg(n, fmt)
{
	static char buf[128];

	switch (fmt) {
	case ARABIC:
		(void) sprintf(buf, "%d", n);
		break;
	case ROMAN:
		return roman_numerals(n, LOWER);
	case ROMANCAPS:
		return roman_numerals(n, UPPER);
	case ALPHA:
		return alpha_num(n, LOWER);
	case ALPHACAPS:
		return alpha_num(n, UPPER);
	default:
		(void) sprintf(buf, "%0*d", fmt, n);
	}
	return buf;
}


static reg *lookup(name)
char *name;
{
	register reg *rp;

	for (rp = regs; rp; rp = rp->next)
		if (streq(rp->name, name))
			return rp;
	return (reg *)0;
}

static reg *getreg(name)
char *name;
{
	register reg *rp;

	rp = lookup(name);
	if (rp == (reg *)0)
		return addreg(name);
	return rp;
}

static reg *addreg(name)
char *name;
{
	register reg *rp;

	rp = (reg *)alloc(sizeof(*rp), "register structure");
	rp->name = mk_string(name);
	rp->value = rp->incr = 0;
	rp->fmt = ARABIC;
	rp->next = regs;
	return regs = rp;	/* reg added to beginning of list */
}
