#include	"defs.h"
#include	"input.h"
#include	"instruct.h"
#include	"print.h"
#include	<setjmp.h>

#define	CODELEN	BUFSIZE
#define	DATALEN	128

instr	code[CODELEN];		/* generated code */
static	instr	*cp;		/* code pointer */

float	t_data[DATALEN];	/* temp data space */
static	instr	t_datalen;	/* number of temp data */
static	int	dflt_scale;	/* default number scaling */
static	bool	no_errs_allowed;

static	jmp_buf	jbuf;

static	void	put();
static	void	expr(), relexpr(), add(), mult(), sfact(), ufact();
static	int	scale_unit();
static	void	synerr();

bool parse(scale, no_errs)
bool no_errs;
{
	cp = code;
	t_datalen = 0;
	dflt_scale = scale;
	no_errs_allowed = no_errs;
	if (setjmp(jbuf))
		return F;
	expr();
	put(EOCODE);
	return T;
}

static void expr()
{
	instr c;

	relexpr();
	loop {
		switch (peekch()) {
		case '&':
			c = AND;
			break;
		case '|':
			c = OR;
			break;
		default:
			return;
		}
		(void) getch();
		relexpr();
		put(c);
	}
}

static void relexpr()
{
	instr c;

	add();
	loop {
		switch (peekch()) {
		case '<':
			c = LT;
			(void) getch();
			if (peekch() == '=') {
				c = LE;
				(void) getch();
			}
			break;
		case '>':
			c = GT;
			(void) getch();
			if (peekch() == '=') {
				c = GE;
				(void) getch();
			}
			break;
		case '!':
		case '=':
			c = getch() == '!' ? NE : EQ;
			if (peekch() == '=')
				(void) getch();
			else if (c == NE)
				synerr("bad operator");
			break;
		default:
			return;
		}
		add();
		put(c);
	}
}

static void add()
{
	instr c;

	mult();
	loop {
		switch (peekch()) {
		case '+':
			c = ADD;
			break;
		case '-':
			c = SUB;
			break;
		default:
			return;
		}
		(void) getch();
		mult();
		put(c);
	}
}

static void mult()
{
	instr c;

	sfact();
	loop {
		switch(peekch()) {
		case '*':
			c = MUL;
			break;
		case '/':
			c = DIV;
			break;
		case '%':
			c = MOD;
			break;
		default:
			return;
		}
		(void) getch();
		sfact();
		put(c);
	}
}

static void sfact()
{
	instr c;

	switch (peekch()) {
	case '-':
		c = UMIN;
		(void) getch();
		break;
	default:
		c = 0;
	}
	ufact();
	if (c)
		put(c);
}

static void ufact()
{
	char c;
	char numbuf[40], *np;

	c = peekch();
	if (c == '(') {
		(void) getch();
		expr();
		if (getch() != ')')
			synerr("missing ')'");
	} else if ((c >= '0' && c <= '9') || c == '.') {
		int scale;
		extern bool isfloat();
		extern double atof();

		np = numbuf;
		*np++ = getch();
		loop {
			c = getch();
			if ((c < '0' || c > '9') && c != '.' && c != 'e')
				break;
			*np++ = c;
		}
		*np = EOS;
		if ((scale = scale_unit(c)) == 0) {
			scale = dflt_scale;
			ungetch(c);
		}
		if (isfloat(numbuf) == F)
			synerr("badly formed number");
		if (t_datalen == DATALEN-1)
			synerr("data space overflow");
		put(LOAD);
		t_data[t_datalen] = (float)(atof(numbuf) * scale);
		put(t_datalen++);
	} else {
		(void) getch();
		synerr("bad character");
	}
}

static int scale_unit(c)
char c;
{
	extern int current_vs(), current_ps();

	switch (c) {
	case 'u':
	case 'p': return 1;
	case 'm': return current_ps();
	case 'n': return current_ps() / 2;
	case 'v': return current_vs();
	case 'i': return 72;
	case 'c': return 28;	/* approximative */
	}
	return 0;
}


static void put(instruct)
instr	instruct;
{
	if (cp >= &code[CODELEN])
		synerr("code table overflow");
	*cp++ = instruct;
}


static void synerr(s)
char *s;
{
	char ebuf[80];

	if (no_errs_allowed == T) {
		(void) sprintf(ebuf, "%s in expression", s);
		warn(ebuf);
	}
	longjmp(jbuf, 1);
}
