#include	<stdio.h>
#include	<time.h>
#include	"defs.h"
#include	"output.h"
#include	"print.h"
#include	"string.h"
#include	"patchlevel.h"
#include	"../config.h"

#define	VERSION	"2.4"
#define	DATE	"30th June 1994"

#define set(v, n)	if (argv[0][n] == EOS) {\
			if (argc > 1) { argc--; v(*++argv); } else error = T;\
			} else v(&argv[0][n])


/* Underlay text defaults (no underlay by default) */
char	*underlay_txt = "DRAFT";	/* the text to be underlayed */
int	underlay_ps = 90;		/* the point-size of the text */
bool	underlay_solid = F;		/* outline text by default */

static	char	*pfile = NULLSTR;	/* standard PS prolog file name */
static	char	*upfile = NULLSTR;	/* underlay text PS prolog file name */
static	char	*mfile = NULLSTR;	/* macro package name */

static	void	pr_prolog();
static	void	pr_ulyprolog();
static	void	setup_date();
static	void	copyfile();
static	void	read_macros();
static	void	usage();

extern	void	format();
extern	char	*getenv();
extern	void	exit();


main(argc, argv)
char *argv[];
{
	FILE *fp;
	bool prolog = T, underlay = F, error = F;
	extern bool warning, landscape;
	extern void init_ps(), init_fi(), init_register(), do_pagelist();
	extern int atoi();

	while (argc > 1) { 
		if (argv[1][0] != '-' || argv[1][1] == EOS)	/* filename */
			break;
		argc--; argv++;
		switch (argv[0][1]) {
		case 'w':	/* print warning messages on errors */
			warning = T;
			break;
		case 'l':	/* landscape mode */
			landscape = T;
			break;
		case 'n':
			switch (argv[0][2]) {
			case 'f':	/* no fill mode */
				init_fi(0); break;
			case 'a':	/* fill but no adjust */
				init_fi(1); break;
			case EOS:	/* don't output prolog */
				prolog = F; break;
			default:
				error = T;
			}
			break;
		case 'p':	/* use named PS prolog file */
			set(pfile = , 2);
			break;
		case 'm':	/* use macro package */
			set(mfile = , 2);
			break;
		case 'r':	/* initialise register */
			set(init_register, 2);
			break;
		case 'o':	/* set output page list */
			set(do_pagelist, 2);
			break;
		case 'u':	/* underlay text options */
			switch (argv[0][2]) {
			case 's':	/* solid underlay text */
				underlay_solid = T;
				/* FALLTHROUGH */
			case 'o':	/* outline underlay text */
			case EOS:	/* outline text by default */
				underlay = T;
				break;
			case 't':	/* set the underlay text itself */
				set(underlay_txt = , 3);
				underlay = T;
				break;
			case 'p':	/* use named underlay prolog file */
				set(upfile = , 3);
				break;
			case 'z':	/* set the underlay text point size */
				set(underlay_ps = atoi, 3);
				break;
			default:
				error = T;
			}
			break;
		case 'v':	/* print version number and die */
			pr("psfmt : version %s patch %d, dated %s\n",
				VERSION, PATCH, DATE);
			exit(0);
		default:
			error = T;
		}
		if (error == T)
			usage("bad option '%s'\n", argv[0]);
	}

	/*
	 * generate necessary prolog
	 */
	if (prolog == T)
		pr_prolog();
	if (underlay == T)
		pr_ulyprolog();
	else
		underlay_ps = 0;	/* use as flag to indicate no underlay */
	do_pagelist(NULLSTR);
	setup_date();
	init_ps();

	/*
	 * format all the input files
	 */
	if (mfile)
		read_macros();
	if (argc > 1) {
		while (--argc) {
			if (streq(*++argv, "-"))
				format(stdin, "");
			else {
				if ((fp = fopen(*argv, "r")) == NULL)
					ferr("can't open %s\n", *argv);
				format(fp, *argv);
			}
		}
	} else
		format(stdin, "");

	/*
	 *	generate the trailer
	 */
	out_br();
	pr("\n@end\n");
	return 0;
}


static void pr_prolog()
{
	FILE *pfp;

	if (pfile == NULLSTR && (pfile = getenv("PSFMTPROLOG")) == NULLSTR)
		pfile = PSFPROLOG;
	if ((pfp = fopen(pfile, "r")) == NULL)
		ferr("cannot open prolog file '%s' for reading\n", pfile);
	copyfile(pfp);
}


static void pr_ulyprolog()
{
	FILE *upfp;

	if (upfile == NULLSTR && (upfile = getenv("PSFMTUPROLOG")) == NULLSTR)
		upfile = PSFUPROLOG;
	if ((upfp = fopen(upfile, "r")) == NULL)
		ferr("cannot open underlay text prolog file '%s' for reading\n", upfile);
	copyfile(upfp);
}


static void setup_date()
{
	long date;
	struct tm *t;
	static char *abbday[7] =
		{"Sun","Mon","Tue","Wed","Thu","Fri","Sat"};
	static char *fullday[7] =
		{"Sunday","Monday","Tuesday","Wednesday",
		"Thursday","Friday","Saturday"};
	static char *abbmon[12] =
		{"Jan","Feb","Mar","Apr","May","Jun",
		"Jul","Aug","Sep","Oct","Nov","Dec"};
	static char *fullmon[12] =
		{"January","February","March","April","May","June","July",
		"August","September","October","November","December"};
	extern void set_ireg(), set_istr();
#ifdef __hpux
	date = time((time_t *)0);
	t = localtime( (time_t *)&date);
#else
	date = time((long *)0);
	t = localtime(&date);
#endif
	set_ireg("dw", t->tm_wday + 1);
	set_ireg("dy", t->tm_mday);
	set_ireg("mo", t->tm_mon + 1);
	set_ireg("yr", t->tm_year);
	set_istr("dw", abbday[t->tm_wday]);
	set_istr("day", fullday[t->tm_wday]);
	set_istr("mo", abbmon[t->tm_mon]);
	set_istr("month", fullmon[t->tm_mon]);
}


static void copyfile(fp)
FILE *fp;
{
	char buf[BUFSIZE];
	int n;

	while ((n = fread(buf, 1, BUFSIZE, fp)) > 0)
		(void) fwrite(buf, 1, n, stdout);
	(void) fclose(fp);
}


static void read_macros()
{
	FILE *mfp;
	char file[BUFSIZE], *path;

	if (*mfile == EOS)
		return;
	if ((mfp = fopen(mfile, "r")) != NULL)
		(void) strcpy(file, mfile);
	else {
		if ((path = getenv("PSFMTMACDIR")) == NULLSTR)
			path = PSFMACDIR;
		(void) sprintf(file, "%s/psf.%s", path, mfile);
		if ((mfp = fopen(file, "r")) == NULL)
			ferr("cannot open macro file '%s' for reading\n", file);
	}
	format(mfp, file);
}

static char *option_notes[] = {
	"-w		produce warning messages on errors",
	"-l		output pages in landscape mode",
	"-n		don't output the PostScript prolog",
	"-p<prolog>	use the file <prolog> as the PostScript prolog",
	"-nf		start off in no-fill mode",
	"-na		start off in fill mode without adjusting",
	"-m<macros>	initially read the file <macros>",
	"-o<pagelist>	print only pages in <pagelist>",
	"-r<reg>		initialise register <reg> with value 0",
	"-r<reg>=<val>	initialise register <reg> with value <val>",
	"-v		print psfmt's version number and exit",
	"-u or -uo	print outlined underlay text on each page",
	"-us		print solid (gray) underlay text on each page",
	"-ut<str>	use the string <str> as the underlay text",
	"-uz<ps>		use the pointsize <ps> for underlay text",
	"-up<prolog>	use the file <prolog> as the underlay prolog",
	(char *)0
};

static void usage(f, a)
char *f, *a;
{
	char *opts = "[-w] [-l] [-n|-p<prolog>] [-nf|-na] [-m<macros>]\n\
	[-o<pagelist>] [-r<reg>[=<val>] [-v]\n\
	[-u[o|s]] [-ut<str>] [-uz<ps>] [-up<prolog>]";
	register char **op = option_notes;

	(void) fprintf(stderr, f, a);
	(void) fprintf(stderr, "usage : psfmt %s [file ...]\n\n", opts);
	while (*op)
		(void) fprintf(stderr, "\t%s\n", *op++);
	exit(1);
}
