#include	<stdio.h>
#include	"../config.h"
#include	"defs.h"
#include	"input.h"
#include	"interpol.h"
#include	"scalenum.h"
#include	"string.h"
#include	"output.h"
#include	"print.h"


/* ENVIRONMENT HANDLING */
#define	NENVIRONS	5
#define	CURVAL		0
#define	OLDVAL		1

#define	CUR(x)		env[elev].x[CURVAL]
#define	OLD(x)		env[elev].x[OLDVAL]
#define	CURE(x)		env[elev].x
#define EXCH(x)		t=CUR(x); CUR(x)=OLD(x); OLD(x)=t
#define EXCHS(x)	ts=CUR(x); CUR(x)=OLD(x); OLD(x)=ts
#define SET(x, val)	CUR(x) = val
#define SAVE(x)		OLD(x) = CUR(x)
#define	SETE(x, val)	CURE(x) = val

static	int	t;		/* temp use in EXCH */
static	char	*ts;		/* temp use in EXCHS */

#if	A4PAPER
/* default top~10.3"; bottom=1"; offset~1.2"; linelen=6"; boxmargin=0.5" */
/* landscape: top=7"; bottom=1"; offset=1"; linelen~9.7" */
#define	DEFTOP		742
#define	DEFBOT		72
#define	DEFOFFSET	86
#define	DEFLINELEN	432
#define	DEFBOXMARG	36
#define	DEFBOTBOXMARG	22
#define	DEFLSTOP	504
#define	DEFLSBOT	72
#define	DEFLSOFFSET	72
#define	DEFLSLINELEN	698
#define	PAGEWID		596
#define	PAGEHGHT	840
#else
/* default top~9.6"; bottom=1"; offset~1.44"; linelen=6"; boxmargin=0.5" */
/* landscape: top=7.5"; bottom=1"; offset=1"; linelen=9" */
#define	DEFTOP		692
#define	DEFBOT		72
#define	DEFOFFSET	104
#define	DEFLINELEN	432
#define	DEFBOXMARG	36
#define	DEFBOTBOXMARG	22
#define	DEFLSTOP	540
#define	DEFLSBOT	72
#define	DEFLSOFFSET	72
#define	DEFLSLINELEN	648
#define	PAGEWID		612
#define	PAGEHGHT	792
#endif

static struct env {
	/* variables that remember previous value */
	char	*font[2];	/* font abbreviation */
	int	ps[2];		/* point size */
	int	vs[2];		/* vertical spacing */
	int	in[2];		/* indentation */
	int	off[2];		/* page offset */
	int	ll[2];		/* line length */
	int	tl[2];		/* title length */
	/* variables unique in environment */
	bool	fi;		/* set if filling lines */
	bool	ad;		/* set if also justifying */
	int	tin;		/* temporary indent */
} env[NENVIRONS] = {
	{ {"T","T"}, {12,12}, {14,14}, {0,0}, {DEFOFFSET,DEFOFFSET},
	  {DEFLINELEN,DEFLINELEN}, {DEFLINELEN,DEFLINELEN}, T, T, 0 },
	{ {"C","C"}, {10,10}, {12,12}, {0,0}, {DEFOFFSET,DEFOFFSET},
	  {DEFLINELEN,DEFLINELEN}, {DEFLINELEN,DEFLINELEN}, F, F, 0 },
	{ { NULLSTR } },	/* envs 3 to 5 defined on first use */
	{ { NULLSTR } },
	{ { NULLSTR } }
};
static	int	elev = 0;		/* current environment level */
static	int	estack[NENVIRONS - 1];	/* pushed environments stack */
static	int	*esp = estack - 1;	/* environment stack pointer */

/* FORMATTING CONTROL */
#define	FONT	01
#define	PS	02
#define	VS	04
#define	IN	010
#define	TIN	020
#define	OFF	040
#define	LL	0100
#define	TL	0200
#define	FI	0400
#define	ADJ	01000

static	int	changed = 0;		/* set if format style changed */
static	bool	multicol = F;		/* set when multi-columning */
int	centre = 0;			/* centre next n lines */
bool	fill = T;			/* fill lines to start with */
bool	nobreak;			/* set on no-break requests */
bool	partline = F;			/* set if partial line output */
bool	landscape = F;			/* set for landscape mode */

/* USER MODIFIABLE FORMATTING CONTROL */
static	int	top = DEFTOP;		/* top of printed page */
static	int	bottom = DEFBOT;	/* bottom of printed page */
static	int	pagelen = DEFTOP-DEFBOT;/* (printed) page length */
static	int	pagewid = DEFLINELEN;	/* (printed) page width */
static	int	bm_top = DEFBOXMARG;	/* top box margin */
static	int	bm_bot = DEFBOTBOXMARG;	/* bottom box margin */
static	int	bm_left = DEFBOXMARG;	/* left box margin */
static	int	bm_rght = DEFBOXMARG;	/* right box margin */
static	int	pageno = 1;		/* current page no. */
static	bool	dobox = T;		/* print box */
static	bool	dopageno = T;		/* print page no. */
static	int	display = -1;		/* set to spacing when in a display */

/* FUNCTION DECLARATIONS */
void		set_prev_font();	/* set previous font and ps */
void		do_break();		/* force a line break */
void		do_changes();		/* do any request formatting changes */
void		init_ps();		/* initialise PostScript environment */
char		*current_font();	/* return buf with current font & ps */
int		current_vs();		/* return current vert. spacing */
static	void	change_ps();		/* change point size */
static	void	set_family_font();	/* set font [RIBX] in current family */
static	void	change_font();		/* change font */
static	void	change_inline_font();	/* change font within a line */
static	void	change_str_font();	/* change font only for next string */
static	void	pop_environment();	/* go back to previous environment */
static	bool	push_environment();	/* push down new environment */
static	void	num_arg_to_bool();	/* convert numeric arg to bool value */
static	int	cmp_env();		/* compare environments 0 & 1 */
static	void	set_iregs();		/* set internal reg values */
extern	char	*lookup_font();		/* font lookup routine */
extern	char	*lookup_famfont(); 	/* font within family lookup routine */
extern	void	outword();		/* output interpolated word */


void set_local_ps()		/* \sn or \s(nn */
{
	if (num.sign == ABS && num.val == 0) {
		EXCH(ps);
		return;
	}
	SAVE(ps);
	set_num(&CUR(ps), 1);
}

void set_ps()			/* .ps [p] [v] */
{
	if (peeknb() == EOS) {
		EXCH(ps);
		EXCH(vs);
		change_ps(1);
		return;
	}
	scale_num(1, F);
	if (num.err == F) {
		SAVE(ps);
		set_num(&CUR(ps), 1);
		change_ps(0);
	}
	if (peeknb() == EOS)
		return;
	scale_num(1, T);
	if (num.err == T)
		RET_WARN("illegal vertical spacing with '.ps'");
	SAVE(vs);
	set_num(&CUR(vs), 1);
	changed |= VS;
	set_iregs(VS);
}

static void change_ps(vs)
{
	if (partline == T)
		change_inline_font();
	else
		changed |= PS;
	if (vs)
		changed |= VS;
	set_iregs(vs ? PS|VS : PS);
}

void set_local_font(f, p2)	/* \ff or \f(ff or \f{f[:p]} */
char *f;
bool p2;
{
	char *ft;

	if (streq(f, "P")) {
		EXCHS(font);
		EXCH(ps);
		return;
	}
	if (f[1] == EOS && strchr("RIBX", f[0]) != NULLSTR)
		ft = lookup_famfont(f[0], CUR(font));
	else
		ft = lookup_font(f);
	if (ft == NULLSTR)
		RET_WARN("unknown font in \\f sequence");
	SAVE(font);
	SAVE(ps);
	SET(font, ft);
	if (p2 == T)
		set_num(&CUR(ps), 1);
}

void set_font()			/* .ft [f [p] [v]] */
{
	register char *ft;

	if (peeknb() == EOS) {
		set_prev_font();
		return;
	}
	ft = lookup_font(getwd());
	if (ft == NULLSTR)
		RET_WARN("unknown fontname with '.ft'");
	SAVE(font);
	SAVE(ps);
	SET(font, ft);
	if (peeknb() != EOS)
		set_ps();
	change_font();
}

void set_temp_font()		/* .tf f str */
{
	register char *ft;

	if (peeknb() == EOS)
		RET_WARN("no arguments given to '.tf'");
	ft = lookup_font(getwd());
	if (ft == NULLSTR)
		RET_WARN("unknown fontname with '.tf'");
	change_str_font(ft, 1);
}

void set_roman()		/* .r [str] */
{
	set_family_font('R');
}

void set_italic()		/* .i [str] */
{
	set_family_font('I');
}

void set_bold()			/* .b [str] */
{
	set_family_font('B');
}

void set_bolditalic()		/* .x [str] */
{
	set_family_font('X');
}

void set_prev_font()		/* .p */
{
	EXCHS(font);
	EXCH(ps);
	change_font();
}

static void set_family_font(f)
char f;
{
	register char *ft;

	ft = lookup_famfont(f, CUR(font));
	if (ft == NULLSTR)
		RET_WARN("current font not part of a family");
	if (peeknb() == EOS) {
		SAVE(font);
		SAVE(ps);
		SET(font, ft);
		change_font();
		return;
	}
	change_str_font(ft, 0);
}

static void change_font()
{
	if (partline == T)
		change_inline_font();
	else
		changed |= FONT;
	set_iregs(FONT);
}

static void change_inline_font()
{
	char tfont[35];

	(void) sprintf(tfont, "[%s]", current_font());
	out_str(tfont);
}

static void change_str_font(font, tf_req)
char *font;
{
	char tfont[40], *wp;

	get_string();
	if (*string == EOS)
		RET_WARN(tf_req ? "no or null string argument given to '.tf'"
			 : "null string argument given to .r, .i, .b or .x");
	do_changes();
	partline = T;
	(void) sprintf(tfont, "[%d %s]", CUR(ps), font);
	wp = interpolate(string, tfont, current_font(), 0);
	if (fill == T)
		outword(wp, 1, 1);
	else
		out_line(wp, NOT(concatenate));
}

void set_fill()			/* .fi */
{
	do_break();
	out_fi(CURE(ad));
	SETE(fi, fill = T);
	set_iregs(FI);
}

void set_nofill()		/* .nf */
{
	do_break();
	out_nf();
	SETE(fi, fill = F);
	set_iregs(FI);
}

void set_adjust()		/* .ad */
{
	do_break();
	SETE(ad, T);
	out_adj(CURE(ad));
	set_iregs(ADJ);
}

void set_noadjust()		/* .na */
{
	do_break();
	SETE(ad, F);
	out_adj(CURE(ad));
	set_iregs(ADJ);
}

void set_display()		/* .ds [sp] [f [p] [v]] */
{
	int cur_in = CUR(in);

	if (display != -1)
		RET_WARN("already in display");
	do_break();
	if (push_environment(1) == F)
		return;
	scale_num(1, F);
	if (num.err == F)
		set_num(&display, 0);
	else
		display = 1;	/* space & current indent */
	if (peeknb() != EOS)
		set_font();
	if (display < 2) {
		SAVE(in);
		SET(in, cur_in);
	}
	changed |= cmp_env();
	if (display & 01)
		out_sp(CUR(vs)/2);
	do_changes();
}

void end_display()		/* .de */
{
	if (display == -1)
		RET_WARN("not in a display");
	do_break();
	pop_environment();
	if (display & 01)
		out_sp(CUR(vs)/2);
	display = -1;
}

void set_title()		/* .tl 'left'centre'right' */
{
	char *sptr[3];
	int i;

	if (peeknb() == EOS)
		RET_WARN("no title string given to '.tl'");
	do_changes();
	get_string();
	if (split_string(sptr) == F)
		RET_WARN("null string argument given to '.tl'");
	for (i = 2; i >= 0; i--)
		pr("%s", interpolate(sptr[i], NULLSTR, NULLSTR, 0));
	out_tit();
}

void set_centre()		/* .ce [n] */
{
	if (peeknb() == EOS) {
		centre = 1;
		return;
	}
	scale_num(1, T);
	if (num.err == T || num.sign == MIN)
		RET_WARN("bad argument given to '.ce'");
	set_num(&centre, 0);
}

void set_indent()		/* .in [n] */
{
	do_break();
	if (peeknb() == EOS) {
		EXCH(in);
		changed |= IN;
		changed &= ~TIN;
		return;
	}
	scale_num(72, T);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.in'");
	SAVE(in);
	set_num(&CUR(in), 0);
	changed |= IN;
	changed &= ~TIN;
}

void set_list_indent(n)
{
	SAVE(in);
	CUR(in) = n;
	changed |= IN;
}

void set_temp_indent()		/* .ti [n] */
{
	do_break();
	if (peeknb() == EOS) {
		changed |= TIN;
		return;
	}
	scale_num(72, T);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.ti'");
	CURE(tin) = CUR(in);
	set_num(&CURE(tin), 0);
	changed |= TIN;
}

void set_offset()		/* .po [n] */
{
	do_break();
	if (peeknb() == EOS) {
		EXCH(off);
		changed |= OFF;
		set_iregs(OFF);
		return;
	}
	scale_num(72, T);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.po'");
	SAVE(off);
	if (num.val == 0)
		SET(off, DEFOFFSET);
	else
		set_num(&CUR(off), 0);
	changed |= OFF;
	set_iregs(OFF);
}

void set_linelen()		/* .ll [n] */
{
	if (peeknb() == EOS) {
		EXCH(ll);
		changed |= LL;
		set_iregs(LL);
		return;
	}
	scale_num(72, T);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.ll'");
	SAVE(ll);
	if (num.val == 0)
		SET(ll, DEFLINELEN);
	else
		set_num(&CUR(ll), 0);
	changed |= LL;
	set_iregs(LL);
}

void set_titlelen()		/* .lt [n] */
{
	if (peeknb() == EOS) {
		EXCH(tl);
		changed |= TL;
		return;
	}
	scale_num(72, T);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.lt'");
	SAVE(tl);
	if (num.val == 0)
		SET(tl, DEFLINELEN);
	else
		set_num(&CUR(tl), 0);
	changed |= TL;
}

void set_multicol()		/* .mc [n] [w] */
{
	int ncols = 2, gutter = 24;

	if (multicol == T)
		RET_WARN("already doing multi-column output");
	do_changes();
	out_br();
	if (peeknb() == EOS) {
		multicol = T;
		out_mcol(ncols, gutter);
		return;
	}
	scale_num(1, F);
	if (num.err == F)
		set_num(&ncols, 2);
	if (peeknb() != EOS) {
		scale_num(72, T);
		if (num.err == T)
			RET_WARN("bad gutter width given to '.mc'");
		set_num(&gutter, 0);
	}
	multicol = T;
	out_mcol(ncols, gutter);
}

void set_singlecol()		/* .sc */
{
	if (multicol == F)
		RET_WARN("not doing multi-column output");
	do_changes();
	out_br();
	multicol = F;
	out_scol();
}

void next_column()		/* .nc [b] */
{
	bool pageonoverflow = F;

	if (multicol == F)
		RET_WARN("not doing multi-column output");
	do_break();
	if (peeknb() == EOS) {
		out_ncol(pageonoverflow);
		return;
	}
	scale_num(1, T);
	if (num.err == F)
		num_arg_to_bool(&pageonoverflow);
	out_ncol(pageonoverflow);
}

void set_environment()		/* .ev [n] */
{
	int e = 0;

	if (peeknb() == EOS) {
		pop_environment();
		return;
	}
	scale_num(1, T);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.ev.");
	set_num(&e, 1);
	if (e < 1 || e > NENVIRONS)
		RET_WARN("bad environment number given to '.ev'");
	if (env[--e].font[0] == NULLSTR) /* if environment never been used, */
		env[e] = env[0];	 /* copy over the 'global' one */
	(void) push_environment(e);
}

static void pop_environment()
{
	int tmp = *esp;

	if (esp < estack)
		RET_WARN("not currently in a nested environment");
	*esp = elev;
	elev = tmp;
	do_changes();
	changed |= cmp_env();
	esp--;
}

static bool push_environment(e)
{
	if (esp == &estack[NENVIRONS - 2]) {
		warn("environments currently fully nested");
		return F;
	}
	*++esp = elev;
	elev = e;
	do_changes();
	changed |= cmp_env();
	return T;
}

void set_tabstops()		/* .ta [n ... ] */
{
	char tbuf[512];
	int s = 0;

	tbuf[0] = EOS;
	if (peeknb() == EOS) {
		int l = CUR(ll);

		for (s = 36; s < l; s += 36)
			(void) sprintf(tbuf, "%s%d ", tbuf, s);
		out_stops(tbuf);
		return;
	}
	do {
		scale_num(72, T);
		if (num.err == T || num.sign == MIN) {
			warn("bad tab stop value with '.ta'");
			break;
		}
		set_num(&s, s);
		(void) sprintf(tbuf, "%s%d ", tbuf, s);
	} while (peeknb() != EOS);
	out_stops(tbuf);
}

void set_top()			/* .top n */
{
	if (peeknb() == EOS)
		RET_WARN("no argument given to '.top'");
	scale_num(72, F);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.top'");
	set_num(&top, 0);
	out_top(top);
}

void set_pagelen()		/* .pl n */
{
	if (peeknb() == EOS)
		RET_WARN("no argument given to '.pl'");
	scale_num(72, F);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.pl'");
	set_num(&pagelen, 0);
	out_plen(pagelen);
	bottom = top - pagelen;
	out_bot(bottom);
}

void set_pagewid()		/* .pw n */
{
	if (peeknb() == EOS)
		RET_WARN("no argument given to '.pw'");
	scale_num(72, F);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.pw'");
	set_num(&pagewid, 36);
	out_pwid(pagewid);
}

void set_pageno()		/* .pn n */
{
	if (peeknb() == EOS)
		RET_WARN("no argument given to '.pn'");
	scale_num(1, T);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.pn'");
	set_num(&pageno, ANY);
	out_pageno(pageno);
}

void set_box()			/* .box [b] [p] */
{
	if (peeknb() == EOS) {
		dobox = dopageno = T;
		out_box(dobox);
		out_pn(dopageno);
		return;
	}
	scale_num(1, F);
	if (num.err == F) {
		num_arg_to_bool(&dobox);
		out_box(dobox);
	}
	if (peeknb() != EOS) {
		scale_num(1, T);
		if (num.err == F) {
			num_arg_to_bool(&dopageno);
			out_pn(dopageno);
		}
	}
}

static void num_arg_to_bool(b)
bool	*b;
{
	int n = *b == T ? 1 : 0;

	set_num(&n, ANY);
	*b = n > 0 ? T : F;
}

void set_boxmarg()		/* .bm [t] [b] [l] [r] */
{
	if (peeknb() == EOS) {
		bm_top = bm_bot = bm_left = bm_rght = DEFBOXMARG;
		out_bm(bm_top, bm_bot, bm_left, bm_rght);
		return;
	}
	scale_num(72, F);
	if (num.err == F)
		set_num(&bm_top, 0);
	scale_num(72, F);
	if (num.err == F)
		set_num(&bm_bot, 0);
	scale_num(72, F);
	if (num.err == F)
		set_num(&bm_left, 0);
	scale_num(72, F);
	if (num.err == F)
		set_num(&bm_rght, 0);
	out_bm(bm_top, bm_bot, bm_left, bm_rght);
}

void set_escape()		/* .ec [n] */
{
	int argval = do_escape_sequ == T ? 1 : 0;

	if (peeknb() == EOS) {
		do_escape_sequ = NOT(do_escape_sequ);
		return;
	}
	scale_num(1, T);
	if (num.err == T)
		RET_WARN("non-numeric argument given to '.ec'");
	set_num(&argval, 0);
	do_escape_sequ = argval != 0 ? T : F;
}

void do_break()			/* .br */
{
	if (partline == T && nobreak == F)
		out_br();
	partline = F;
}

void do_breakpage()		/* .bp */
{
	do_break();
	out_page();
}

void do_space()			/* .sp [|][n] */
{
	int sp;
	bool move = F;

	sp = CUR(vs) / 2;
	if (peeknb() != EOS) {
		if (peeknb() == '|') {
			(void) getnb();
			move = T;
		}
		scale_num(sp, T);
		if (num.err == T)
			RET_WARN("non-numeric argument given to '.sp'");
		if (move == T) {
			do_break();
			if (num.sign == MIN)
				out_y(bottom + num.val);
			else
				out_y(top - num.val);
			partline = F;
			return;
		}
		sp = num.sign == MIN ? -num.val : num.val;
	}
	do_break();
	out_sp(sp);
}

void do_need()			/* .ne [n] */
{
	int need = CUR(vs);

	if (peeknb() != EOS) {
		scale_num(need, T);
		if (num.err == T || num.sign == MIN)
			RET_WARN("bad argument given to '.ne'");
		need = num.val;
	}
	do_break();
	out_need(need);
}


void do_changes()
{
	if (changed == 0)
		return;
	if ((changed & FONT) || (changed & PS))
		out_font(CUR(font), CUR(ps));
	if (changed & VS)
		out_vs(CUR(vs));
	if ((changed & OFF) || (changed & LL))
		out_rmarg(CUR(off)+CUR(ll));
	if ((changed & IN) || (changed & OFF))
		out_x(CUR(in)+CUR(off));
	if (changed & OFF)
		out_po(CUR(off));
	if (changed & LL) {
		out_ll(CUR(ll));
		out_setl();
	}
	if (changed & TL)
		out_tl(CUR(tl));
	if (changed & FI) {
		if (CURE(fi) == T)
			out_fi(CURE(ad));
		else
			out_nf();
	} else if (changed & ADJ)
		out_fi(CURE(ad));
	if (changed & TIN)
		out_tx(CURE(tin)+CUR(off));
	pr("\n");
	set_iregs(changed);
	changed = 0;
}

void init_ps()
{
	extern int underlay_ps;
	extern bool underlay_solid;
	extern char *underlay_txt;

	if (landscape == T) {
		top = DEFLSTOP;
		bottom = DEFLSBOT;
		env[0].off[0] = env[0].off[1] = env[1].off[0] = env[1].off[1] =
			DEFLSOFFSET;
		env[0].ll[0] = env[0].ll[1] = env[1].ll[0] = env[1].ll[1] =
		env[0].tl[0] = env[0].tl[1] = env[1].tl[0] = env[1].tl[1] =
			DEFLSLINELEN;
		pagelen = DEFLSTOP - DEFLSBOT;
		pagewid = DEFLSLINELEN;
		pr("/@setp{%d 0 translate 90 rotate}bind def\n", PAGEWID);
	} else
		pr("/@setp{}def\n");
	out_top(top);
	out_bot(bottom);
	out_plen(pagelen);
	out_pwid(CUR(ll));
	out_bm(bm_top, bm_bot, bm_left, bm_rght);
	out_box(dobox);
	out_pn(dopageno);
	out_pageno(pageno);
	out_po(CUR(off));
	pr("\ncurrentdict/x known not{/x po def/y top def/@np? @T def}if");
	pr("\n@setp go\n");
	if (underlay_ps)
		pr("/@ut(%s)def/@up %d def/@us? %s def @sul @pul\n",
			underlay_txt, underlay_ps, bool_str(underlay_solid));
	changed |= FONT | VS | LL | TL | FI;
	do_changes();
	if (fill == F)	/* fake definitions for save/restore handling */
		pr("/a @T def/@sz 0 def/@wf? @F def/@kf 0 def\n");
	pr("%% end of prolog %%\n");
	pr("/@snap save def\n");
	set_iregs(FONT|PS|VS|IN|OFF|LL|FI|ADJ);
}

void init_fi(n)
{
	if (n)	/* no adjust */
		SETE(ad, F);
	else	/* no fill */
		SETE(fi, fill = F);
}

void ch_source()		/* .so [file] */
{
	FILE *fp;
	char *file;

	if (peeknb() == EOS) {
		fp = stdin;
		file = "stdin";
	} else {
		file = getwd();
		if ((fp = fopen(file, "r")) == NULL)
			RET_WARN("couldn't open source file for reading");
	}
	do_changes();
	pushinfile(fp, file);
}


char *current_font()
{
	static char fontstr[30];

	(void) sprintf(fontstr, "[%d %s]", CUR(ps), CUR(font));
	return fontstr;
}

int current_vs()
{
	return CUR(vs);
}

int current_ps()
{
	return CUR(ps);
}

int current_in()
{
	return CUR(in);
}

int abs_x(n)
{
	return CUR(off) + n;
}

#if NEEDED
int abs_y(n)
{
	return top - n;
}
#endif

static int cmp_env()
{
#define CMP(x)		env[elev].x[CURVAL] != env[p].x[CURVAL]
#define VCMP(x)		env[elev].x != env[p].x
#define SCMP(x)		!(streq(env[elev].x[CURVAL], env[p].x[CURVAL]))
	register int c = 0, p = *esp;

	if (SCMP(font))	c |= FONT;
	if (CMP(ps))	c |= PS;
	if (CMP(vs))	c |= VS;
	if (CMP(in))	c |= IN;
	if (CMP(off))	c |= OFF;
	if (CMP(ll))	c |= LL;
	if (CMP(tl))	c |= TL;
	if (VCMP(fi))	c |= FI;
	if (VCMP(ad))	c |= ADJ;
	fill = CURE(fi);
	set_iregs(c);
	return c;
}

static void set_iregs(n)
{
	extern void set_ireg(), set_istr();

	if (n & FONT)	set_istr(".f", CUR(font));
	if (n & PS)	set_ireg(".s", CUR(ps));
	if (n & VS)	set_ireg(".v", CUR(vs));
	if (n & IN)	set_ireg(".i", CUR(in));
	if (n & OFF)	set_ireg(".o", CUR(off));
	if (n & LL)	set_ireg(".l", CUR(ll));
	if (n & TL)	set_ireg(".t", CUR(tl));
	if (n & FI)	set_ireg(".u", CURE(fi) == T ? 1 : 0);
	if (n & ADJ)	set_ireg(".j", CURE(ad) == T ? 1 : 0);
}
