#include	"defs.h"
#include	"input.h"
#include	"name.h"
#include	"line.h"
#include	"string.h"
#include	"print.h"
#include	"alloc.h"

#define	STRING	0
#define	MACRO	1

bool	copymode = F;

typedef struct S {
	char	*name;	/* macro/string name */
	int	type;	/* macro or string flag */
	itext	s_m;	/* the actual macro lines or string */
	struct S *next;
} mstr;

static	mstr	*mstrs = (mstr *)0;
static	mstr	*call_macro;

static	mstr	*lookup();
static	mstr	*getmstr();
static	mstr	*addmstr();
static	void	free_text();


void set_istr(name, val)	/* set internal string */
char *name, *val;
{
	getmstr(name, STRING)->s_m.string = mk_string(val);
}

void set_string()		/* .str str [val] */
{
	register mstr *mp;

	if (peeknb() == EOS)
		RET_WARN("no arguments given to '.str'");
	mp = getmstr(trim(getwd()), STRING);
	copymode = T;
	get_string();
	copymode = F;
	if (*string == EOS)
		return;
	mp->s_m.string = mk_string(string);
}

void set_macro()		/* .def mac\n<anything>\n...\n.. */
{
	char linebuf[4096];
	register mstr *mp;
	register char *lbp;
	register line *lp, *lq;
	bool done = F;

	if (peeknb() == EOS)
		RET_WARN("no argument given to '.def'");
	mp = getmstr(trim(getwd()), MACRO);
	copymode = T;
	while (done == F && readline(">> ") == T) {
		if (streq(getln(ALL), "..")) {
			(void) strcpy(linebuf, ".PArgs");
			done = T;
		} else
			for (lbp = linebuf; (*lbp++ = getch()) != EOS; );
		lp = (line *)alloc(sizeof(*lp), "macro");
		lp->text = mk_string(linebuf);
		lp->next = (line *)0;
		if (mp->s_m.lines == (line *)0)
			lq = mp->s_m.lines = lp;
		else
			lq = lq->next = lp;
	}
	copymode = F;
}

void rm_strmac()		/* .rm str/mac ... */
{
	register mstr *mp, *mq;

	if (peeknb() == EOS)
		RET_WARN("no argument given to '.rm'");
	do {
		if ((mp = lookup(trim(getwd()))) == (mstr *)0) {
			warn("non-existent macro/string name given to '.rm'");
			continue;
		}
		for (mq = mstrs; mq; mq = mq->next)
			if (mq->next == mp)
				break;
		if (mq == (mstr *)0)	/* first on list */
			mstrs = mp->next;
		else
			mq->next = mp->next;
		free(mp->name);
		free_text(mp, -1);
		free((char *)mp);
	} while (peeknb() != EOS);
}

void interp_str()		/* \*s or \*(ss or \*{str} */
{
	char c;
	int n = 1;
	mstr *mp;

	c = getch();
	if (c == '(' || c =='{')
		n = c == '(' ? 2 : 0;
	else
		ungetch(c);
	mp = lookup(trim(getname(n)));
	if (mp == (mstr *)0)
		RET_WARN("interpolation of undefined string");
	if (mp->s_m.string)
		pushinstr(mp->s_m.string);
}

static void exec_string()
{
	itext i;

	if (call_macro->s_m.string == NULLSTR)
		return;
	i.string = call_macro->s_m.string;
	pushinline(i, call_macro->name, 1);
}

static void exec_macro()
{
	itext i;
	extern void push_args();

	if (call_macro->s_m.lines == (line *)0)
		return;
	push_args();
	i.lines = call_macro->s_m.lines;
	pushinline(i, call_macro->name, 0);
}

void (*macro(req))()
char *req;
{
	register mstr *mp;

	for (mp = mstrs; mp; mp = mp->next)
		if (streq(mp->name, req)) {
			call_macro = mp;
			return mp->type == STRING ? exec_string : exec_macro;
		}
	return (void (*)())0;
}


static mstr *lookup(name)
char *name;
{
	register mstr *mp;

	for (mp = mstrs; mp; mp = mp->next)
		if (streq(mp->name, name))
			return mp;
	return (mstr *)0;
}

static mstr *getmstr(name, type)
char *name;
{
	register mstr *mp;

	mp = lookup(name);
	if (mp == (mstr *)0)
		return addmstr(name, type);
	free_text(mp, type);
	return mp;
}

static mstr *addmstr(name, type)
char *name;
{
	register mstr *mp;

	mp = (mstr *)alloc(sizeof(*mp), "string/macro structure");
	mp->name = mk_string(name);
	mp->type = type;
	if (type == STRING)
		mp->s_m.string = NULLSTR;
	else
		mp->s_m.lines = (line *)0;
	mp->next = mstrs;
	return mstrs = mp;	/* add to beginning of list */
}

static void free_text(mp, newtype)
register mstr *mp;
{
	if (mp->type == STRING) {
		if (mp->s_m.string)
			free(mp->s_m.string);
	} else {
		register line *lp, *lq;

		for (lp = mp->s_m.lines; lp; lp = lq) {
			lq = lp->next;
			free(lp->text);
			free((char *)lp);
		}
	}
	switch (newtype) {
	case STRING:
		mp->s_m.string = NULLSTR;
		mp->type = STRING;
		break;
	case MACRO:
		mp->s_m.lines = (line *)0;
		mp->type = MACRO;
	}
}
